#include <assert.h>
#include <dlfcn.h>
#include <msgpack.h>
#include <stdlib.h>
#include <threads.h>
#include <time.h>
#include <zmq.h>

typedef struct sqlcxp
{
  unsigned short fillen;
  char *filnam;
} sqlcxp;

typedef struct sqltvn
{
  unsigned char *tvnvsn;
  unsigned short tvnvsnl;
  unsigned char *tvnnm;
  unsigned short tvnnml;
  unsigned char *tvnsnm;
  unsigned short tvnsnml;
} sqltvn;

typedef struct sqladts
{
  unsigned int adtvsn;
  unsigned short adtmode;
  unsigned short adtnum;
  sqltvn adttvn[1];
} sqladts;

typedef struct sqltdss
{
  unsigned int tdsvsn;
  unsigned short tdsnum;
  unsigned char *tdsval[1];
} sqltdss;

typedef struct sqlexd
{
  unsigned long sqlvsn;
  unsigned int arrsiz;
  unsigned int iters;
  unsigned int offset;
  unsigned short selerr;
  unsigned short sqlety;
  unsigned int occurs;
  const short *cud;
  unsigned char *sqlest;
  const char *stmt;
  sqladts *sqladtp;
  sqltdss *sqltdsp;
  unsigned char **sqphsv;
  unsigned long *sqphsl;
  int *sqphss;
  short **sqpind;
  int *sqpins;
  unsigned long *sqparm;
  unsigned long **sqparc;
  unsigned short *sqpadto;
  unsigned short *sqptdso;
  unsigned int sqlcmax;
  unsigned int sqlcmin;
  unsigned int sqlcincr;
  unsigned int sqlctimeout;
  unsigned int sqlcnowait;
  int sqfoff;
  unsigned int sqcmod;
  unsigned int sqfmod;
  unsigned int sqlpfmem;
  unsigned char *sqhstv[1];
  unsigned long sqhstl[1];
  int sqhsts[1];
  short *sqindv[1];
  int sqinds[1];
  unsigned long sqharm[1];
  unsigned long *sqharc[1];
  unsigned short sqadto[1];
  unsigned short sqtdso[1];
} sqlexd;

typedef struct varchar
{
  unsigned short len;
  unsigned char arr[];
} varchar;

typedef void (*sqlcxt_t) (void **, unsigned int *, struct sqlexd *,
                          const struct sqlcxp *);

enum exec_opcode
{
  sql_open = 267,
  sql_for_x_fetch = 270,
  sql_close = 271,
  sql_prepare = 273,
  sql_declare = 275,
  sql_describe = 276,
  sql_execute = 277,
  sql_for_x_execute = 279,
  sql_execute_immediate = 280,
  sql_connect = 283,
  sql_commit_work = 285,
  sql_commit_work_release = 286,
  sql_rollback_work = 287,
  sql_context_allocate = 314,
  sql_context_free = 315,
  sql_enable_threads = 316
};

struct oraprof
{
  sqlcxt_t sqlcxt;
  void *socket;
  mtx_t mutex;
} static oraprof = { 0, 0 };

__attribute__ ((unused)) void
sqlcxt (void **ctx, unsigned int *sqlctx, struct sqlexd *sqlstm,
        const struct sqlcxp *sqlfpn)
{
  if (!oraprof.sqlcxt)
    {
      oraprof.sqlcxt = dlsym (RTLD_NEXT, "sqlcxt");
      assert (oraprof.sqlcxt != 0);

      void *zctx = zmq_ctx_new ();
      assert (zctx);

      oraprof.socket = zmq_socket (zctx, ZMQ_PUSH);
      assert (oraprof.socket);

      int rc = -1;

      const char *addr = getenv ("ORAPROF_ENDPOINT");
      if (addr == 0)
        {
          fprintf (stderr,
                   "[W] ORAPROF_ENDPOINT not set, profiling disabled\n");
        }
      else
        {
          rc = zmq_bind (oraprof.socket, addr);
          if (rc != 0)
            {
              fprintf (stderr, "[E] %s\n", strerror (errno));
            }
          else
            {
              if (mtx_init (&oraprof.mutex, mtx_plain) != thrd_success)
                {
                  fprintf (stderr, "[E] Failed to initialize mutex");
                  rc = -1;
                }
            }
        }

      if (rc != 0)
        {
          zmq_close (oraprof.socket);
          zmq_ctx_destroy (zctx);
          mtx_destroy (&oraprof.mutex);
          oraprof.socket = 0;
        }
    }

  if (sqlstm != 0 && oraprof.socket != 0)
    {
      struct timespec ts;
      clock_gettime (CLOCK_REALTIME, &ts);

      short opcode = sqlstm->cud[sqlstm->offset + 6];

      msgpack_sbuffer sbuf;
      msgpack_sbuffer_init (&sbuf);

      msgpack_packer packer;
      msgpack_packer_init (&packer, &sbuf, msgpack_sbuffer_write);

      msgpack_pack_array (&packer, 10);
      msgpack_pack_unsigned_long (&packer, ctx != 0 ? (unsigned long)*ctx : 0);
      msgpack_pack_unsigned_long (&packer, thrd_current ());
      msgpack_pack_short (&packer, opcode);

      switch (opcode)
        {
        case sql_open:
        case sql_close:
        case sql_declare:
        case sql_describe:
        case sql_execute:
        case sql_commit_work:
        case sql_commit_work_release:
        case sql_rollback_work:
        case sql_context_allocate:
        case sql_context_free:
        case sql_enable_threads:
          msgpack_pack_nil (&packer);
          break;

        case sql_for_x_fetch:
        case sql_for_x_execute:
          {
            char buf[16];
            size_t len = snprintf (buf, 16, "%u", sqlstm->iters);
            msgpack_pack_str (&packer, len);
            msgpack_pack_str_body (&packer, buf, len);
            break;
          }

        case sql_prepare:
        case sql_execute_immediate:
        case sql_connect:
          if (sqlstm->sqhstl[0] > 0)
            {
              varchar *vc = (varchar *)sqlstm->sqhstv[0];
              msgpack_pack_str (&packer, vc->len);
              msgpack_pack_str_body (&packer, vc->arr, vc->len);
            }
          else
            {
              size_t len = strlen ((const char *)sqlstm->sqhstv[0]);
              msgpack_pack_str (&packer, len);
              msgpack_pack_str_body (&packer, sqlstm->sqhstv[0], len);
            }
          break;

        default:
          if (sqlstm->stmt != 0)
            {
              size_t len = strlen (sqlstm->stmt);
              msgpack_pack_str (&packer, len);
              msgpack_pack_str_body (&packer, sqlstm->stmt, len);
              break;
            }
          else
            {
              msgpack_pack_nil (&packer);
            }
          break;
        }

      struct timespec tic;
      clock_gettime (CLOCK_MONOTONIC, &tic);

      oraprof.sqlcxt (ctx, sqlctx, sqlstm, sqlfpn);

      struct timespec toc;
      clock_gettime (CLOCK_MONOTONIC, &toc);

      msgpack_pack_long (&packer, ts.tv_sec);
      msgpack_pack_long (&packer, ts.tv_nsec);
      msgpack_pack_long (&packer, tic.tv_sec);
      msgpack_pack_long (&packer, tic.tv_nsec);
      msgpack_pack_long (&packer, toc.tv_sec);
      msgpack_pack_long (&packer, toc.tv_nsec);

      mtx_lock (&oraprof.mutex);
      int rc = zmq_send (oraprof.socket, sbuf.data, sbuf.size, ZMQ_DONTWAIT);
      mtx_unlock (&oraprof.mutex);

      if (rc == -1)
        {
          fprintf (stderr, "[E] %s\n", strerror (errno));
        }

      msgpack_sbuffer_destroy (&sbuf);
    }
  else
    {
      oraprof.sqlcxt (ctx, sqlctx, sqlstm, sqlfpn);
    }
}
